<?php

namespace App\Entity;

//use App\Repository\NoteRepository;
//use Doctrine\ORM\Mapping as ORM;

//#[ORM\Entity(repositoryClass: NoteRepository::class)]
class Note
{
    //#[ORM\Id]
    //#[ORM\GeneratedValue]
    //#[ORM\Column]

    private int    $id;
    private int    $mod;
    private array  $terms = [];
    private string $profile;
    private array  $tags = [];
    private string $model;
    // Maybe these doesn't make sense to keep but leaving it here just in
    // case for handiness' sake 
    private array  $fields = [];
    private array  $cardIds;

    const HIGHLIGHT_PATTERN = '/<span\s+([^>]*)>(.*?)<\/span>/i';
    const HIGHLIGHT_ATTR_KANJI = 'style="color: rgb(255, 78, 8);"';

    public function getId(): int
    {
        return $this->id;
    }

    public function hasTerm(string $kanji): bool
    {
        foreach ($this->terms as $term) {
            assert($term instanceof Term);

            if ($term->kanji == $kanji) return true;
        }

        return false;
    }

    public static function fromAnki(array $noteInfo): self
    {
        $note = new self();

        [
            'noteId'    => $note->id,
            'mod'       => $note->mod,
            'profile'   => $note->profile,
            'tags'      => $note->tags,
            'modelName' => $note->model,
            'cards'     => $note->cardIds,
        ] = $noteInfo;

        $note->fields = array_map(fn($x) => $x['value'], $noteInfo['fields']);

        // Set VocabKanji field
        //$vocabKanji = explode('｜', $note->fields['VocabKanji']);
        $note->terms = self::parseVocabDef($note->fields['VocabDef']);

        // If not defined, find them from the highlighted parts in the sentence
        if (empty($note->terms)) {
            // 1. Get all spans in the text
            preg_match_all(
                self::HIGHLIGHT_PATTERN,
                $note->fields['SentKanji'],
                $matches,
                PREG_SET_ORDER,
            );
            // 2. Check the ones that match with the kanji color
            foreach ($matches as $match) {
                if ($match[1] === self::HIGHLIGHT_ATTR_KANJI) {
                    $term = new Term();
                    $term->kanji = mb_trim($match[2]);
                    $term->definitionEn = null;
                    $term->definitionJp = null;
                    $note->terms[] = $term;
                }
            }
        }

        // Set to null whatever is null
        $readings = array_map(
            fn($x) => in_array($x, ['_', '＿', '']) ? null : $x,
            explode('｜', $note->fields['VocabFurigana']),
        );

        // Set readings from furigana field
        foreach ($note->terms as $key => &$term) {
            if (null === $term->getReading()) {
                if (null !== ($readings[$key] ?? null)) {
                    $term->kanji .= '[' . $readings[$key] . ']';
                }
            }
        }

        return $note;
    }

    public function toAnki(): array
    {
        return [
            'id' => $this->id,
            'fields' => [
                'VocabKanji' => join('｜', array_map(
                    fn(Term $x) => $x->getKanji(),
                    $this->terms,
                )),
                'VocabFurigana' => join('｜', array_map(
                    fn(Term $x) => $x->getReading() ?? '＿',
                    $this->terms,
                )),
                'VocabDef' => join("<br>\n", array_map(
                    fn(Term $x) => $x->toAnkiVocabDef(),
                    $this->terms,
                )),
            ],
        ];
    }

    public static function parseVocabDef(string $vocabDef): array
    {
        if (mb_trim($vocabDef) == "") return [];

        $terms = [];

        foreach (preg_split('|<br ?/?>|', $vocabDef) as $line) {
            $term = Term::fromVocabDefLine(strip_tags($line));
            if (null === $term) dd("error parsing term", $line);
            $terms[] = $term;
        };

        return $terms;
    }

    public function getCreatedAt(): \DateTimeImmutable
    {
        $timestamp = ceil($this->id / 1000);
        return \DateTimeImmutable::createFromFormat('U', $timestamp);
    }

    public function getUpdatedAt(): \DateTimeImmutable
    {
        return \DateTimeImmutable::createFromFormat('U', $this->mod);
    }
}
